from abc import ABC, abstractmethod

from flloat.base.Alphabet import Alphabet
from flloat.syntax.ldlf import LDLfFormula
from gym.spaces import Discrete

from pythomata.base.Symbol import Symbol
from typing import Set

from rltg.agents.feature_extraction import FeatureExtractor
from rltg.logic.PartialAutomatonSimulator import PartialAutomatonSimulator
from rltg.logic.RewardAutomaton import RewardAutomaton
from rltg.logic.RewardAutomatonSimulator import RewardAutomatonSimulator


class TemporalEvaluator(ABC):
    def __init__(self, goal_feature_extractor:FeatureExtractor, alphabet:Set[Symbol], formula:LDLfFormula, reward,
                 gamma=0.99, on_the_fly=False):
        self.goal_feature_extractor = goal_feature_extractor
        self.alphabet = Alphabet(alphabet)
        self.formula = formula
        self.on_the_fly = on_the_fly
        if not on_the_fly:
            self._automaton = RewardAutomaton._fromFormula(alphabet, formula, reward, gamma=gamma)
            self.simulator = RewardAutomatonSimulator(self._automaton)
        else:
            self.dfaotf = self.formula.to_automaton(alphabet, on_the_fly=True)
            self.simulator = PartialAutomatonSimulator(self.dfaotf, self.alphabet, reward, gamma=gamma)

    @abstractmethod
    def fromFeaturesToPropositional(self, features, action, *args, **kwargs) -> Set[Symbol]:
        raise NotImplementedError

    def update(self, action, state):
        """update the automaton.
        :param action: the action to reach the state
        :param state:  the new state of the MDP
        :returns (new_automaton_state, reward)"""
        features = self.goal_feature_extractor(state)
        propositional = self.fromFeaturesToPropositional(features, action)
        reward = self.simulator.make_transition(propositional)
        return self.simulator.get_current_state(), reward

    def get_state(self):
        return self.simulator.get_current_state()

    def get_state_space(self):
        if not self.on_the_fly:
            return Discrete(len(self.simulator.state2id))
        else:
            # estimate the state space size.
            # the estimate MUST overestimate the true size.
            # TODO: do you really return a magic number?
            return Discrete(100)


    def reset(self):
        self.simulator.reset()

    def is_failed(self):
        return self.simulator.is_failed()

    def is_true(self):
        return self.simulator.is_true()



import random

from gym import Space

from rltg.agents.exploration_policies.ExplorationPolicy import ExplorationPolicy


class RandomPolicy(ExplorationPolicy):

    def __init__(self, action_space:Space, epsilon=0.1, epsilon_start=None, decaying_steps=1):
        super().__init__()
        self.action_space = action_space
        self.exploration_steps = decaying_steps

        if epsilon_start is None:
            self.epsilon = epsilon
            self.epsilon_start = None
        else:
            self.epsilon_start = epsilon_start
            self.epsilon_end = epsilon
            self.epsilon = self.epsilon_start
            self.epsilon_decay_step = (epsilon_start - epsilon) / self.exploration_steps

    def update(self, *args):
        if self.epsilon_start is not None:
            if self.epsilon > self.epsilon_end:
                self.epsilon -= self.epsilon_decay_step
            else:
                self.epsilon = self.epsilon_end

    def explore(self, *args):
        action_id = None
        if random.random() < self.epsilon:
            action_id =self.action_space.sample()
        return action_id

    def reset(self, *args):
        pass



"""Base class for every RL agent."""

import _pickle as pickle

from rltg.agents.brains.Brain import Brain
from rltg.agents.exploration_policies import ExplorationPolicy
from rltg.agents.feature_extraction import RobotFeatureExtractor


class RLAgent(object):

    def __init__(self, sensors: RobotFeatureExtractor, exploration_policy:ExplorationPolicy, brain:Brain, eval:bool=False):
        """
        :param sensors:             feature extractor from the environment observation.
        :param exploration_policy:  A wrapper class to the exploration policy during the learning process.
        :param brain:               The Brain, a wrapper for the actual learning algorithm.
        :param eval:                If eval is True, prevent the brain to learn (testing mode).
                                    otherwise work as usual (learning mode)
                                    default: False

        :raises ValueError          if the space of the feature extractor output is different from the space of
                                    the brain input.
        """
        if brain.observation_space and sensors.output_space != brain.observation_space:
            raise ValueError("space dimensions are not compatible.")
        self.sensors = sensors
        self.exploration_policy = exploration_policy
        self.brain = brain
        self.eval = eval

    def set_eval(self, eval:bool):
        """Setter method for "eval" field."""
        self.eval = eval

    def act(self, state, best_action=False):
        """Extract the features and call _act
        :param state:       the state from which the agent makes a move.
        :raises ValueError: if It the state is not contained into sensors.input_space
        :returns the chosen action
        """
        features = self.sensors(state)
        return self._act(features, best_action=best_action)


    def _act(self, features, best_action=False):
        """From features, choose an action.
        :param features: the state from which the agent makes a move.
        :returns the chosen action
        """
        action = None
        if not self.eval and not best_action:
            # if in evaluation mode, do not explore.
            action = self.exploration_policy.explore()
        if action is None:
            # if the exploration policy does not specify any action
            # or if we are either in evaluation mode or we need the optimal action,
            # then ask the action to the policy
            action = self.brain.choose_action(features, optimal=best_action)
        return action

    def observe(self, state, action, reward, state2):
        """Called at each observation. """
        features_1 = self.sensors(state)
        features_2 = self.sensors(state2)
        self._observe(features_1, action, reward, features_2)

    def _observe(self, features, action, reward, features2):
        self.brain.observe(features, action, reward, features2)

    def replay(self):
        if not self.eval:
            self.brain.learn()

    def update(self):
        """Called at the end of each iteration.
        It MUST be called only once."""
        if not self.eval:
            self.brain.update()
            self.exploration_policy.update()

    def reset(self):
        """Called at the end of each episode.
        It MUST be called only once."""
        if not self.eval:
            self.brain.reset()
            self.exploration_policy.reset()

    def save(self, filepath):
        with open(filepath + "/exploration_policy.dump", "wb") as fout:
            self.exploration_policy.reset()
            pickle.dump(self.exploration_policy, fout)
        with open(filepath + "/brain.dump", "wb") as fout:
            self.brain.reset()
            pickle.dump(self.brain, fout)
        with open(filepath + "/sensors.dump", "wb") as fout:
            pickle.dump(self.sensors, fout)

    def load(self, filepath):
        with open(filepath + "/exploration_policy.dump", "rb") as fin:
            self.exploration_policy = pickle.load(fin)
        with open(filepath + "/brain.dump", "rb") as fin:
            self.brain = pickle.load(fin)
        with open(filepath + "/sensors.dump", "rb") as fin:
            self.sensors = pickle.load(fin)



from copy import copy
from flloat.base.Alphabet import Alphabet
from flloat.base.Symbol import Symbol
from flloat.syntax.ldlf import LDLfFormula
from pythomata.base.DFA import DFA
from typing import Set


class RewardAutomaton(DFA):
    def __init__(self, dfa:DFA, alphabet:Alphabet, f:LDLfFormula, reward, gamma=0.99):
        super().__init__(
            dfa.alphabet,
            dfa.states,
            dfa.initial_state,
            dfa.accepting_states,
            dfa.transition_function
        )
        self._dfa = dfa
        self.alphabet = alphabet
        self.f = f
        self.reward = reward

        self.reachability_levels,   \
        self.max_level,             \
        self.failure_states = self._compute_levels()

        self.gamma = gamma

    @staticmethod
    def _fromFormula(alphabet:Set[Symbol], f:LDLfFormula, reward, gamma=0.99):
        dfa = f.to_automaton(alphabet, determinize=True, minimize=True)
        return RewardAutomaton(dfa, dfa.alphabet, f, reward, gamma=gamma)

    def get_formula_reward(self):
        return self.reward

    def get_immediate_reward(self, q, q_prime):
        phi = self.potential_function
        if q_prime in self.failure_states:
            return -self.reward/self.max_level
            # return 0


        r = self.gamma * phi(q_prime) - phi(q)
        if q != q_prime:
            # implementation trick: do not scale reward if we are in the same state
            r /= self.max_level
            r *= self.reward

        return r

    def complete(self):
        return RewardAutomaton(self._dfa.complete(), self.alphabet, self.f, self.reward,gamma=self.gamma)

    def minimize(self):
        return RewardAutomaton(self._dfa.minimize(), self.alphabet, self.f, self.reward, gamma=self.gamma)

    def trim(self):
        return RewardAutomaton(self._dfa.trim(), self.alphabet, self.f, self.reward,gamma=self.gamma)

    def potential_function(self, q):
        return self.max_level - self.reachability_levels[q]

    def _compute_levels(self):
        level = 0
        state2level = {final_state: level for final_state in self._dfa.accepting_states}

        z_current = set()
        z_next = set(self._dfa.accepting_states)
        while z_current != z_next:
            level += 1
            z_current = z_next
            z_next = copy(z_current)
            for s in self._dfa.states:
                if s in z_current:
                    continue
                for a in self.transition_function.get(s, []):
                    next_state = self.transition_function[s][a]
                    if next_state in z_current:
                        z_next.add(s)
                        state2level[s] = level

        z_current = z_next

        # levels for failure state (i.e. that cannot reach a final state)
        failure_states = set()
        for s in filter(lambda x: x not in z_current, self._dfa.states):
            state2level[s] = level
            failure_states.add(s)

        max_level = level - 1
        if max_level==0: max_level = 1
        return state2level, max_level, failure_states




"""
        The potential function works as the following:
        - Compute the level of reachability for the final state (precomputed):
        >>> level = self.reachability_levels[q]
        The lower, the nearer to any final state (zero means that q is a final state)
        - Make it "the higher, the nearer":
        >>> true_level = self.max_level - level
        - Normalize over the maximum level possible so that
        the nearer to any final state, the higher the ratio (from 0.0 to 1.0)
        >>> level_ratio = true_level / self.max_level
        - Give the partial reward simply by using the ratio just computed;
        >>> state_reward = level_ratio * self.reward
        :param q: the state of the automaton
        :return: the value of the potential function
        """



from abc import abstractmethod
from time import sleep

import cv2


class Renderer(object):
    @abstractmethod
    def update(self, *args, **kwargs):
        raise NotImplementedError

class PixelRenderer(Renderer):

    def __init__(self, width=600, height=600, window_name='obs', delay=1, skip_frame=1, video=False):
        self.window_name = window_name
        self.delay = delay
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(window_name, width, height)

        self.video = video
        if video:
            self.vid = cv2.VideoWriter('demo.avi', cv2.VideoWriter_fourcc(*"XVID"), float(30), (160, 210), False)

        self.counter = 0
        self.skip_frame = skip_frame


    def update(self, env):
        screen = env.render()
        self.counter += 1
        if self.counter%self.skip_frame==0:
            cv2.imshow(self.window_name, screen)
            cv2.waitKey(self.delay)
            if self.video:
                self.vid.write(screen)

    def release(self):
        if self.video:
            self.vid.release()

class PygameRenderer(Renderer):
    def __init__(self, delay=0.1, skip_frame=1):
        self.counter = 0
        self.skip_frame = skip_frame
        self.delay = delay

    def update(self, env):
        self.counter += 1
        if self.counter % self.skip_frame == 0:
            env.render()
            sleep(self.delay)




from flloat.flloat import DFAOTF
from typing import Set, List

from flloat.semantics.pl import PLInterpretation
from flloat.utils import powerset
from pythomata.base.Alphabet import Alphabet
from pythomata.base.DFA import DFA
from pythomata.base.Symbol import Symbol

from rltg.logic.RewardAutomaton import RewardAutomaton
from rltg.logic.RewardAutomatonSimulator import RewardSimulator


class PartialAutomatonSimulator(RewardSimulator):

    def __init__(self, dfaotf:DFAOTF, alphabet:Alphabet, reward, gamma=0.99):
        self.episode = 0
        self.it = 0

        self.dfaotf   = dfaotf
        self.alphabet = Alphabet({PLInterpretation(set(sym)) for sym in powerset(alphabet.symbols)})
        self.reward   = reward
        self.gamma    = gamma
        self.dfaotf.reset()
        initial_state = self.dfaotf.cur_state

        self.id2state = {0: initial_state}
        self.state2id = {initial_state: 0}

        self.states = {0}
        self.initial_state = 0
        self.transition_function = {}
        self.final_states = set()
        self.failure_states = set()

        # a list of (old_state, label, new_state) to keep track of the sequence of states and labels
        self.trace = []
        self.changed = False

        dfa = DFA(alphabet, frozenset(self.states), self.initial_state, frozenset(self.final_states), self.transition_function)
        # keep the current and the last automaton in order to apply Dynamic Reward Shaping
        self._automaton = RewardAutomaton(dfa, alphabet, dfaotf.f, reward, gamma=gamma)
        self._automaton_prime = self._automaton

        self.visited_states = {self.initial_state}

    def reset(self):
        self.dfaotf.reset()
        self.visited_states = {self.initial_state}

        self.episode += 1
        self.it = 0
        # (old state, label, new state) triple
        # for o, i, n in self.trace:
        #     self._update_from_transition(o, i, n)
        # self.trace = []

    def make_transition(self, s:Set[Symbol]):
        self.it+=1
        i = PLInterpretation(s)
        old_state = self.dfaotf.cur_state
        self.dfaotf.make_transition(i)

        new_state = self.dfaotf.cur_state
        self._update_from_transition(old_state, i, new_state)

        if len(self.final_states) > 0:
            # if the goal state has been discovered: least-fixpoint reward works!
            # this is to manage cases when
            if self.state2id[new_state] in self._automaton_prime.failure_states and not self.is_failed():
                return self.reward/10
            else:
                return self.get_immediate_reward(self.state2id[old_state], self.state2id[new_state])
        else:
            return self.get_partial_reward(old_state, new_state)


    def get_partial_reward(self, old_state, new_state):
        if self.is_true():
            return self.reward
        if self.is_failed():
            return -self.reward/10
        if not self.is_failed() and old_state != new_state and new_state not in self.visited_states:
            # give an optimistic reward to help exploration
            return self.reward/10
        return 0

    def get_immediate_reward(self, q, q_prime):
        phi = self._automaton.potential_function
        phi_prime = self._automaton_prime.potential_function

        r = self.gamma * phi_prime(q_prime) - phi(q)
        if q != q_prime:
            # implementation trick: do not scale reward if we are in the same state
            r /= self._automaton_prime.max_level
            r *= self.reward
        if r>11000:
            print(q, q_prime, phi(q), phi_prime(q_prime), r)

        return r

    def is_failed(self):
        return self.dfaotf.cur_state == frozenset()

    def is_true(self):
        return self.dfaotf.is_true()

    def word_acceptance(self, word: List[Symbol]):
        return self.dfaotf.word_acceptance([PLInterpretation(c) for c in word])

    def _update_from_transition(self, old_state, label, new_state):

        old_state_id, old_state_changed = self._add_state(old_state)
        new_state_id, new_state_changed = self._add_state(new_state)

        changed = old_state_changed or new_state_changed

        if changed: print("%s: changed new states"%self.episode, new_state not in self.state2id, old_state not in self.state2id, new_state_id,
              old_state_id)


        if old_state_id not in self.transition_function or label not in self.transition_function[old_state_id]:
            changed = True
            self.transition_function.setdefault(old_state_id, {})[label] = new_state_id
            print("%s: changed transition"%self.episode, old_state_id not in self.transition_function, label not in self.transition_function[old_state_id], old_state_id, label, new_state_id)

        if self.is_failed() and new_state_id not in self.failure_states:
            self.failure_states.add(new_state_id)
            changed = True
            print("%s: changed failed"%self.episode, new_state_id)
        elif self.is_true() and new_state_id not in self.final_states:
            self.final_states.add(new_state_id)
            changed = True
            print("%s: changed true "%self.episode, new_state_id)

        if changed:
            print("why")
            dfa = DFA(self.alphabet, frozenset(self.states), self.initial_state, frozenset(self.final_states),
                  self.transition_function)

            self._automaton = self._automaton_prime
            self._automaton_prime = RewardAutomaton(dfa, self.alphabet, self.dfaotf.f, self.reward, gamma=self.gamma)
            self._automaton_prime.to_dot("automata/%s-%s" % (self.episode, self.it))
        else:
            self._automaton = self._automaton_prime





    def _add_state(self, new_state):
        changed = False
        new_state_id = self.state2id.get(new_state, None)
        if new_state_id is None:
            # first time we see the state
            changed = True
            new_state_id = len(self.states)
            self.states.add(new_state_id)
            self.id2state[new_state_id] = new_state
            self.state2id[new_state] = new_state_id

        return new_state_id, changed

    def get_current_state(self):
        return self.state2id[self.dfaotf.cur_state]



from abc import ABC, abstractmethod

from gym.core import Space


class Brain(ABC):
    """The class which implements the core of the algorithms"""

    def __init__(self, observation_space:Space, action_space:Space):
        """
        :param observation_space: instance of Space or None. If None, it means that the observation space
                                  is not known a priori and so it is not needed for the algorithm.
        :param action_space:      instance of Space.
        """

        self.observation_space = observation_space
        self.action_space = action_space

        self.episode = 0
        self.iteration = 0
        self.episode_iteration = 0
        self.obs_history = []

    @abstractmethod
    def choose_action(self, state, **kwargs):
        """From a state, return the action for the implemented approach.
        e.g. in Q-Learning, select the argmax of the Q-values relative to the 'state' parameter."""
        raise NotImplementedError

    @abstractmethod
    def learn(self):
        """The method performing the learning (e.g. in Q-Learning, update the table)"""
        raise NotImplementedError

    @abstractmethod
    def observe(self, *args):
        """Called at each observation.
        E.g. in args there can be the S,A,R,S' tuple for save it in a buffer
        that will be read in the "learn" method."""
        self.obs_history.append(args)

    def reset(self):
        """action performed at the end of each episode
        Subclasses implementations should call this method.
        """
        self.episode += 1
        self.episode_iteration = 0

    def update(self):
        """action performed at the end of each iteration
        Subclass implementations should call this method.
        """
        self.episode_iteration += 1
        self.iteration += 1




from abc import abstractmethod
from typing import Set

from flloat.semantics.pl import PLInterpretation
from pythomata.base.Simulator import DFASimulator, Simulator
from pythomata.base.Symbol import Symbol
from pythomata.base.utils import Sink

from rltg.logic.RewardAutomaton import RewardAutomaton

class RewardSimulator(Simulator):

    @abstractmethod
    def get_immediate_reward(self, q, q_prime):
        raise NotImplementedError

    @abstractmethod
    def is_failed(self) -> bool:
        raise NotImplementedError


class RewardAutomatonSimulator(DFASimulator, RewardSimulator):
    def __init__(self, dfa:RewardAutomaton):
        super().__init__(dfa)
        self.visited_states = {self.cur_state}

    def reset(self):
        super().reset()
        self.visited_states = {self.cur_state}

    def make_transition(self, s:Set[Symbol]):
        i = PLInterpretation(s)
        old_state = self.cur_state
        super().make_transition(i)
        reward = self.get_immediate_reward(old_state, self.cur_state)
        self.visited_states.add(self.cur_state)

        return reward

    def get_immediate_reward(self, q, q_prime):
        q_id = self.id2state[q]
        q_prime_id = self.id2state[q_prime]
        return self.dfa.get_immediate_reward(q_id, q_prime_id)

    def is_failed(self):
        return self.id2state[self.cur_state] in self.dfa.failure_states

    def get_cur_state(self):
        return self.cur_state



import gym
from gym import Wrapper


class GoalEnvWrapper(Wrapper):
    def __init__(self, env, f):
        super().__init__(env)
        self.f = f

    def step(self, action):
        obs, reward, done, info = super().step(action)
        info["goal"] = self.f(obs, reward, done, info)
        return obs, reward, done, info



from typing import List

import pickle

import os

import re

from rltg.agents.RLAgent import RLAgent
from rltg.agents.feature_extraction import FeatureExtractor, RobotFeatureExtractor, TupleFeatureExtractor
from rltg.agents.brains.Brain import Brain
from rltg.agents.exploration_policies import ExplorationPolicy
from rltg.agents.temporal_evaluator.TemporalEvaluator import TemporalEvaluator
from gym.spaces import Dict, Discrete, Box, Tuple


class TGAgent(RLAgent):
    """Temporal Goal agent"""

    def __init__(self,
                 sensors: RobotFeatureExtractor,
                 exploration_policy:ExplorationPolicy,
                 brain:Brain,
                 temporal_evaluators:List[TemporalEvaluator]):
        assert len(temporal_evaluators)>=1
        super().__init__(sensors, exploration_policy, brain)
        self.temporal_evaluators = temporal_evaluators

        # compute the feature space. It is the cartesian product between
        # the robot feature space output and the automata state space
        # sensors.output_space is expected to be a Tuple.
        robot_feature_space = sensors.output_space
        automata_state_spaces = [temp_eval.get_state_space() for temp_eval in temporal_evaluators]

        # total feature space = (robot feature space, automata 1 state space, automata 2 state space, ... )
        feature_space = Tuple(robot_feature_space.spaces + tuple(automata_state_spaces))

        # Check if the brain has the same input space dimensions,
        # but only if the brain has specified an observation space.
        if brain.observation_space and feature_space != brain.observation_space:
            raise ValueError("The brain has incompatible observation space: {} instead of {}"
                             .format(brain.observation_space, feature_space))

        # map every state from the space (N0, N1, ..., Nn) to a discrete space of dimension N0*N1*...*Nn-1
        self._from_tuple_to_int = TupleFeatureExtractor(feature_space)

    # TODO: allow customization of this component by modularization
    def state_extractor(self, world_state, automata_states: List):
        # the state is a tuple: (features, A1 state, ..., An state)
        state = self.sensors(world_state) + tuple(automata_states)
        collapsed_state = self._from_tuple_to_int(state)
        return collapsed_state

    # TODO: allow customization of this component by modularization
    def reward_extractor(self, world_reward, automata_rewards: List):
        res = world_reward + sum(automata_rewards)
        return res

    def act(self, state, **kwargs):
        sw = state
        automata_states = [te.get_state() for te in self.temporal_evaluators]
        features = self.state_extractor(sw, automata_states)
        return super()._act(features, **kwargs)

    def observe(self, state, action, reward, state2):
        # get the current automata states
        old_states_automata = [te.get_state() for te in self.temporal_evaluators]

        # update the automata states given the new observed state and collect the reward
        states_automata, rewards_automata = zip(*[te.update(action, state2) for te in self.temporal_evaluators])

        old_state  = self.state_extractor(state,  old_states_automata)
        new_state2 = self.state_extractor(state2, states_automata)
        new_reward = self.reward_extractor(reward, rewards_automata)
        super()._observe(old_state, action, new_reward, new_state2)

    def reset(self):
        super().reset()
        for te in self.temporal_evaluators:
            te.reset()

    def save(self, filepath):
        super().save(filepath)
        for idx, te in enumerate(self.temporal_evaluators):
            with open(filepath + "/te%02d.dump"%idx, "wb") as fout:
                te.reset()
                pickle.dump(te, fout)

    def load(self, filepath):
        super().load(filepath)
        temporal_evaluators_files = sorted([f for f in os.listdir(filepath) if re.match(r'te.*\.dump', f)])
        temporal_evaluators = []
        for idx, te_name in enumerate(temporal_evaluators_files):
            with open(filepath + "/%s" % te_name, "rb") as fin:
                te = pickle.load(fin)
                temporal_evaluators.append(te)
        self.temporal_evaluators = temporal_evaluators



from abc import ABC, abstractmethod

import numpy as np
from typing import List

from rltg.agents.temporal_evaluator.TemporalEvaluator import TemporalEvaluator
from rltg.utils.StatsManager import StatsManager


class StoppingCondition(ABC):
    @abstractmethod
    def check_condition(self, *args, stats_manager:StatsManager=None, **kwargs) -> bool:
        raise NotImplementedError

class StatsStoppingCondition(StoppingCondition):
    def __init__(self, window_size=100):
        assert window_size > 0
        self.window_size = window_size

    def check_condition(self, *args, stats_manager: StatsManager = None, **kwargs):
        if stats_manager is None:
            raise Exception

class GoalPercentage(StatsStoppingCondition):
    def __init__(self, window_size=100, min_perc=0.97):
        super().__init__(window_size)
        assert 0.0 < min_perc <= 1.0
        self.min_perc = min_perc

    def check_condition(self, *args, stats_manager:StatsManager=None, **kwargs):
        super().check_condition(*args, stats_manager=stats_manager, **kwargs)

        goal_percentage = np.mean(stats_manager.goals[-self.window_size:])
        return goal_percentage >= self.min_perc

class AvgRewardPercentage(StatsStoppingCondition):
    def __init__(self, window_size=100, target_mean=1.0):
        super().__init__(window_size)
        self.target_mean = target_mean

    def check_condition(self, *args, stats_manager:StatsManager=None, **kwargs):
        super().check_condition(*args, stats_manager=stats_manager, **kwargs)

        avg_reward = np.mean(stats_manager.total_reward_history[-self.window_size:])
        return avg_reward >= self.target_mean


class CheckAutomataInFinalState(StoppingCondition):

    def check_condition(self, *args, temporal_evaluators:List[TemporalEvaluator]=[], **kwargs):
        if len(temporal_evaluators)==0:
            return True
        return all(t.simulator.is_true() for t in temporal_evaluators)




import random
from abc import abstractmethod

import numpy as np
import math

from gym.core import Space
from gym.spaces import Discrete

from rltg.agents.brains.Brain import Brain


class TDBrain(Brain):
    def __init__(self, observation_space:Discrete, action_space:Space, gamma=0.99, alpha=None, nsteps=300):
        super().__init__(observation_space, action_space)

        self.Visits = {}
        self.gamma = gamma
        self.alpha = alpha

        # sparse representation
        self.Q = {}
        self.nsteps = nsteps
        self.obs_history = []

    def choose_action(self, state, optimal=False):
        Q_values = self.Q[state] if state in self.Q else np.zeros((self.action_space.n,))
        argmaxes = np.argwhere(Q_values == Q_values.max()).flatten()
        if optimal:
            # determistic behavior!
            action = argmaxes[0]
        else:
            # allow randomness
            action = random.choice(argmaxes)

        # action = random.choice(argmaxes)
        return action

    def learn(self):
        if self.episode_iteration + 1 < self.nsteps:
            # no enough observations. + 1 for manage the case when nsteps=1 and iteration=0
            return

        tau = self.episode_iteration + 1 - self.nsteps
        self._nsteps_update(self.obs_history[tau: tau + self.nsteps])

    def observe(self, state, action, reward, state2):
        super().observe(state, action, reward, state2)
        # self.obs_history.append((state, action, reward, state2))
        # if len(self.obs_history)==self.nsteps:
            # delete oldest element
            # self.obs_history.pop(0)
        self.incVisits(state, action)

    def reset(self):
        t = self.episode_iteration - self.nsteps
        for i in range(t, self.episode_iteration):
            self._nsteps_update(self.obs_history[i:], last=True)

        self.obs_history = []
        super().reset()


    def _nsteps_update(self, partial_history, last=False):
        if len(partial_history)<1:
            return

        n_reward_return = 0
        N = len(partial_history)
        # gamma^0*r0 + gamma^1*r1 + ... + gamma^N-1*r_n-1
        for i in range(N):
            n_reward_return += partial_history[i][2] * math.pow(self.gamma, i)

        if not last:
            # add the value of the Q function, depending on the type of algorithm (see the getQa method)
            _, _, _, s_tn = partial_history[-1]
            Q_tn = self.getQa(s_tn)
            n_reward_return += math.pow(self.gamma, N) * Q_tn

        s_tau, a_tau, _, _ = partial_history[0]
        delta = n_reward_return - self.Q.get(s_tau, np.zeros(self.action_space.n))[a_tau]

        if s_tau not in self.Q:
            self.Q[s_tau] = np.zeros(self.action_space.n)
        self.Q[s_tau][a_tau] += self.getAlphaVisitsInc(s_tau, a_tau) * delta

    def setVisits(self, x, a, q):
        if not x in self.Visits:
            self.Visits[x] = np.zeros(self.action_space.n)
        self.Visits[x][a] = q

    def incVisits(self, x, a):
        self.setVisits(x,a,self.getVisits(x,a)+1)

    def getVisits(self, x, a):
        if x in self.Visits:
            return self.Visits[x][a]
        else:
            return 0

    def getAlphaVisitsInc(self, x, a):
        if self.alpha:
            return self.alpha

        self.incVisits(x,a)
        s = self.getVisits(x,a)
        try: #TODO debug here
            a = 1.0/float(s)
        except:
            a = 1.0
        #print("visits: %d, a = %.6f" %(s,a))
        return a # math.sqrt(s)

    def getSumVisits(self, x):
        return np.sum(self.Visits[x,:])


    @abstractmethod
    def getQa(self, s):
        raise NotImplementedError



class QLearning(TDBrain):
    def __init__(self, observation_space:Discrete, action_space, gamma=0.99, alpha=None, nsteps=200):
        super().__init__(observation_space, action_space, gamma, alpha, nsteps)

    def getQa(self, s):
        maxQa = max(self.Q.get(s, np.zeros(self.action_space.n)))
        return maxQa


class Sarsa(TDBrain):
    def __init__(self, observation_space:Discrete, action_space, gamma=0.99, alpha=None, nsteps=200):
        super().__init__(observation_space, action_space, gamma, alpha, nsteps)

    def getQa(self, s):
        a = self.choose_action(s)
        q = self.Q.get(s, np.zeros(self.action_space.n))[a]
        return q



import numpy as np
import matplotlib.pyplot as plt

class StatsManager(object):

    def __init__(self, window_size=100):
        self.window_size = window_size
        self.total_reward_history = np.array([], dtype=np.int32)
        self.avg_reward_history = np.array([], dtype=np.float32)
        self.std_reward_history = np.array([], dtype=np.float32)
        self.explored_states_history = np.array([], dtype=np.int32)
        self.goals = np.array([], dtype=np.bool)


    def update(self, n_states, total_reward, goal):
        self.explored_states_history = np.append(self.explored_states_history, n_states)
        self.total_reward_history = np.append(self.total_reward_history, total_reward)
        avg_reward = np.mean(self.total_reward_history[-self.window_size:])
        std_reward = np.std(self.total_reward_history[-self.window_size:])
        self.avg_reward_history = np.append(self.avg_reward_history, avg_reward)
        self.std_reward_history = np.append(self.std_reward_history, std_reward)

        self.goals = np.append(self.goals, goal)


    def print_summary(self, episode_number, step, n_states, total_reward, epsilon, goal):
        w = self.window_size
        if episode_number % 1 ==0:
            print('Episode: {:6d}, Step: {:5d}, Explored States: {:7d}, Total Reward: {:8.2f}, AvgReward: {:8.2f}, StdReward: {:8.2f}, Epsilon: {:05.4f}, Goal: {:5}, GoalPerc: {:05.2f}'
              .format(episode_number, step, n_states, total_reward,
                      np.mean(self.total_reward_history[-w:]),
                      np.std(self.total_reward_history[-w:]),
                      epsilon, str(goal), np.mean(self.goals[-w:])*100)
                  )

    def plot(self):
        plt.figure(1)
        plt.title("Total reward history")
        plt.plot(self.total_reward_history)
        plt.figure(2)
        plt.title("Average total reward history")
        plt.plot(list(self.avg_reward_history))
        plt.figure(3)
        plt.title("Standard deviation total reward history")
        plt.plot(list(self.std_reward_history))
        plt.show()



import shutil

from gym import Env

from rltg.agents.RLAgent import RLAgent
from rltg.agents.TGAgent import TGAgent
from rltg.utils.Renderer import Renderer
from rltg.utils.StatsManager import StatsManager

import os

from rltg.utils.StoppingCondition import CheckAutomataInFinalState, GoalPercentage


class Trainer(object):
    def __init__(self, env:Env, agent:RLAgent, n_episodes=1000,
                 eval=False,
                 resume=False,
                 try_optimal_run=True,
                 renderer:Renderer=None,
                 window_size=100,
                 stopping_conditions=(GoalPercentage(10, 1.0), CheckAutomataInFinalState()),
                 agent_data_dir="agent_data"):
        self.env = env
        self.agent = agent
        self.stopping_conditions = stopping_conditions
        self.n_episodes = n_episodes
        self.resume = resume
        self.eval = eval
        self.try_optimal_run = try_optimal_run
        self.renderer = renderer
        self.window_size = window_size

        if not self.resume:
            shutil.rmtree(agent_data_dir, ignore_errors=True)
            os.mkdir(agent_data_dir)

        self.agent_data_dir = agent_data_dir

        if self.eval:
            self.agent.set_eval(self.eval)

    def main(self):
        agent = self.agent

        num_episodes = self.n_episodes
        last_goal = False
        stats = StatsManager(self.window_size)

        if self.resume:
            agent.load(self.agent_data_dir)

        # Main training loop
        for ep in range(num_episodes):

            # switch between training mode and evaluation mode
            # to check if policy reached is optimal.
            # only when in training mode
            steps, total_reward, goal = self.train_loop(try_optimal=self.eval or (last_goal and self.try_optimal_run))
            last_goal = goal

            stats.update(len(agent.brain.Q), total_reward, goal)
            stats.print_summary(ep, steps, len(agent.brain.Q), total_reward, agent.exploration_policy.epsilon, goal)

            # stopping conditions
            if self.check_stop_conditions(agent, stats):
                break

            agent.reset()
            if not self.eval and ep % 100 == 0:
                agent.save(self.agent_data_dir)

        agent.save(self.agent_data_dir)
        stats.plot()


    def train_loop(self, try_optimal=False):
        env = self.env
        agent = self.agent
        total_reward = 0
        steps = 0

        state = env.reset()
        if self.renderer:
            self.renderer.update(env)

        temporal_evaluators = agent.temporal_evaluators if isinstance(agent, TGAgent) else []

        done, all_te_not_failed, all_te_true = False, True, False
        info = {"goal": False}

        # until the game is not ended and every temporal task is not failed or every temporal task is true
        while not done and all_te_not_failed and not all_te_true:
            action = agent.act(state, best_action=try_optimal)
            state2, reward, done, info = env.step(action)
            agent.observe(state, action, reward, state2)
            agent.replay()

            # add the observed reward (including the automaton reward)
            total_reward += agent.brain.obs_history[-1][2]
            steps += 1

            agent.update()
            state = state2

            if self.renderer:
                self.renderer.update(env)

            all_te_not_failed = all(not t.is_failed() for t in temporal_evaluators)
            all_te_true   = all(t.is_true() for t in temporal_evaluators)

        return steps, total_reward, info["goal"] and all(t.is_true() for t in temporal_evaluators)

    def check_stop_conditions(self, agent, stats):
        temporal_evaluators = agent.temporal_evaluators if isinstance(agent, TGAgent) else []
        if not self.eval and all([s.check_condition(stats_manager=stats, temporal_evaluators=temporal_evaluators)
                                  for s in self.stopping_conditions]):
            return True

        return False



from abc import ABC, abstractmethod


class ExplorationPolicy(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def update(self, *args):
        """Called at each episode iteration"""
        raise NotImplementedError

    @abstractmethod
    def explore(self, *args):
        """Decides what to do as the next action. The actual implementation must be defined into the "_explore" method.
        :returns the action to pick or None if the exploration policy decides to don't explore."""
        raise NotImplementedError

    @abstractmethod
    def reset(self, *args):
        """Called at the end of each episode"""
        raise NotImplementedError



import functools
import operator
import typing
from abc import ABC, abstractmethod

from gym import Space
from gym.spaces import Tuple, Discrete


class FeatureExtractor(ABC):

    def __init__(self, input_space: Space, output_space: Space):
        self.input_space = input_space
        self.output_space = output_space

    def __call__(self, input, **kwargs):
        """Extract features from the input and make sanity check
        of the input and the output dimensions of the abstract method '_extract'"""

        if not self.input_space.contains(input):
            raise ValueError("input space dimensions are not correct.")

        output = self._extract(input, **kwargs)

        if not self.output_space.contains(output):
            raise ValueError("output space dimensions are not correct.")

        return output

    @abstractmethod
    def _extract(self, input, **kwargs):
        """
        Extract the feature from `input` (contained into `self.input_space`).
        The features must be contained into `self.output_space`.
        :param   input:  a state belonging to input_space
        :returns output: a state belonging to output_space
        """
        raise NotImplementedError


class RobotFeatureExtractor(FeatureExtractor):
    """A specialization of the FeatureExtractor used from RLAgent."""
    pass



class IdentityFeatureExtractor(FeatureExtractor):
    def __init__(self, space:Space):
        super().__init__(space, space)

    def _extract(self, input, **kwargs):
        return input


class TupleFeatureExtractor(FeatureExtractor):
    """Collapse a tuple state to an integer, according to the ranges of each dimensions.
    For example:
    consider the scenario with this observation space:
    >>> tuple_space = Tuple((Discrete(10), Discrete(5), Discrete(20)))
    >>> extractor = TupleFeatureExtractor(tuple_space)

    The state (9, 3, 11) which is contained in the observation space, is converted in the following way:
    >>> extracted_feature = (11*10 + 9)*5+3
    >>> extracted_feature
    598
    >>> extractor((9, 3, 11))
    598

    The maximum state (9, 4, 19) is represented in this transformed space as:
    >>> extracted_feature = (19*10 + 9)*5+4
    >>> extracted_feature
    999
    >>> extractor((9, 4, 19))
    999

    which is equal to 20*10*5 - 1

    we sort the space components in descending order of their dimensions, otherwise this method wouldn't work.
    """

    def __init__(self, space:Tuple):
        # assume every subspace is of type Discrete
        # TODO: generalize to different subspaces?
        assert isinstance(space, Tuple) and all(isinstance(s, Discrete) for s in space.spaces)

        # compute output dimension by multiplying all the dimensions
        # i.e.: given a space (N0, N1, N2 ... Nn)
        # the total dimension is: N0*N1*...*Nn
        tot_dim = functools.reduce(operator.mul, map(lambda x: x.n, space.spaces))

        # sort subspaces from the biggest to the smaller, keeping its component id in the input space
        self.id2space_sorted = sorted(enumerate(space.spaces), key=lambda x: -x[1].n)

        super().__init__(space, Discrete(tot_dim))


    def _extract(self, input, **kwargs):

        # new_input = [input[id] for id, _ in self.id2space_sorted]

        # sort the input by descending size of the space components
        state = input[self.id2space_sorted[0][0]]

        # collapse the input into only one number,
        # for doing this, we need to start
        # from the biggest space to the smallest.
        for id, space in self.id2space_sorted[1:]:
            state *= space.n
            state += input[id]

        return state

